#include "ppc.h"
#include "framebuffer.h"
#include <math.h>
#include <float.h>

#include <fstream>
#include <iostream>

PPC::PPC(int _w, int _h, float hfov) : w(_w), h(_h) {

  C = V3(0.0f, 0.0f, 0.0f);
  a = V3(1.0f, 0.0f, 0.0f);
  b = V3(0.0f, -1.0f, 0.0f);
  c[0] = -(float)w/2.0f;
  c[1] = +(float)h/2.0f;
  float tan2 = tanf(hfov/180.0f*3.1415f/2.0f);
  c[2] = -(float)w/2.0f/tan2;

}

V3 PPC::Unproject(int u, int v, float z) {

  if (z == 0.0f)
    return V3(-FLT_MAX, -FLT_MAX, -FLT_MAX);

  float uf = .5f + (float)u;
  float vf = .5f + (float)v;
  V3 ret = C + (a*uf+b*vf+c)/z;
  return ret;

}

bool PPC::Project(V3 P, V3 &projP) {

  M3x3 camM;
  camM.SetColumn(0, a);
  camM.SetColumn(1, b);
  camM.SetColumn(2, c);
  M3x3 invCamM = camM.Invert();
  V3 Q = invCamM*(P-C);

  if (Q[2] < 0.0f)
    return false; // behind head

  projP[0] = Q[0]/Q[2];
  projP[1] = Q[1]/Q[2];
  projP[2] = 1.0f/Q[2];

  return true;

}


void PPC::Visualize(PPC *ppc, float d, FrameBuffer *fb) {

  float scf = d / GetF();
  int psize = 5;
  unsigned int color = 0xFF000000;
  V3 colorv; colorv.SetFromColor(color);
  fb->DrawPoint3D(C, psize, color, ppc);
  V3 a1 = a*scf;
  V3 b1 = b*scf;
  V3 c1 = c*scf;
  fb->DrawSegment3D(C, C+c1, colorv, colorv, ppc);
  fb->DrawSegment3D(C+c1, C+c1+a1*(float)w, colorv, colorv, ppc);
  fb->DrawSegment3D(C+c1+a1*(float)w, C+c1+a1*(float)w+b1*(float)h, 
    colorv, colorv, ppc);
  fb->DrawSegment3D(C+c1+a1*(float)w+b1*(float)h, 
    C+c1+b1*(float)h, 
    colorv, colorv, ppc);
  fb->DrawSegment3D(C+c1+b1*(float)h, C+c1,
    colorv, colorv, ppc);

}

void PPC::Translate(V3 trans) {

  C = C+trans;

}

void PPC::TranslateFW(float step) {

  V3 vd = GetVD();
  Translate(vd*step);

}

V3 PPC::GetVD() {

  V3 ret = a^b;
  ret = ret.Normalized();
  return ret;

}

float PPC::GetF() {

  return c*GetVD();

}

void PPC::ChangeFocalLength(float scalef) {

  float f = GetF();
  float newf = f * scalef;
  V3 tl = C + c;
  V3 newtl = tl + GetVD()*(newf-f);
  c = newtl - C;

}

void PPC::Pan(float thetad) {

  V3 axis = b*-1.0f;
  axis = axis.Normalized();
  a = a.RotateDirectionAboutArbitraryAxis(axis, thetad);
  b = b.RotateDirectionAboutArbitraryAxis(axis, thetad);
  c = c.RotateDirectionAboutArbitraryAxis(axis, thetad);

}

void PPC::Tilt(float thetad) {

  V3 axis = a;
  axis = axis.Normalized();
  a = a.RotateDirectionAboutArbitraryAxis(axis, thetad);
  b = b.RotateDirectionAboutArbitraryAxis(axis, thetad);
  c = c.RotateDirectionAboutArbitraryAxis(axis, thetad);

}

void PPC::Roll(float thetad) {

  V3 axis = GetVD();
  axis = axis.Normalized();
  a = a.RotateDirectionAboutArbitraryAxis(axis, thetad);
  b = b.RotateDirectionAboutArbitraryAxis(axis, thetad);
  c = c.RotateDirectionAboutArbitraryAxis(axis, thetad);

}

void PPC::SetViewGL() {
  
  glViewport(0, 0, w, h);
  // set intrinsics
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  float f = GetF();
  float hither = 1.0f;
  float yon = 1000.0f;
  float scalef = hither / f;
  float wf = a.Length() * w;
  float hf = b.Length() * h;
 // specify rectangle on hither plane and distances to hither and yon planes
 glFrustum (-wf/2.0f*scalef, wf/2.0f*scalef, 
   -hf/2.0f*scalef, hf/2.0f*scalef, hither, yon);

  glMatrixMode (GL_MODELVIEW); // default matrix mode
  // set extrinsics
  V3 look = C + GetVD()*100.0f;
  glLoadIdentity();
  V3 up = (b*-1.0f).Normalized();
  gluLookAt(C[0], C[1], C[2], look[0], look[1], look[2], up[0], up[1], up[2]);

}

void PPC::Save(char *fname) {

  ofstream ofs(fname);
  ofs << a << endl;
  ofs << b << endl;
  ofs << c << endl;
  ofs << C << endl;
  ofs << w << endl;
  ofs << h << endl;
  ofs << "a b c C w h" << endl;
  ofs.close();

}


void PPC::Load(char *fname) {

  ifstream ifs(fname);
  if (ifs.ios::fail()) {
    cerr << "ERROR: cannot open " << fname << endl;
    return;
  }
  ifs >> a ;
  ifs >> b ;
  ifs >> c ;
  ifs >> C ;
  ifs >> w ;
  ifs >> h ;
  ifs.close();

}


float PPC::ComputeError(V3 *pts, V3 *markers, int n) {

  float err = 0.0f;
  for (int i = 0; i < n; i++) {
    V3 projP;
    if (!Project(pts[i], projP))
      return FLT_MAX;
    projP[2] = 0.0f;
    err += (markers[i]-projP).Length();
  }

  return err/(float) n;

}
